from __future__ import annotations

import logging
from inspect import signature
from pathlib import Path

import h5py

from .... import compression, types
from ... import datatype, utils
from ...exceptions import LH5EncodeError
from .array import _h5_write_array
from .scalar import _h5_write_scalar
from .vector_of_vectors import _h5_write_vector_of_vectors

log = logging.getLogger(__name__)


def _h5_write_lgdo(
    obj,
    name,
    lh5_file,
    group="/",
    start_row=0,
    n_rows=None,
    wo_mode="append",
    write_start=0,
    **h5py_kwargs,
):
    assert isinstance(obj, types.LGDO)

    file_kwargs = {
        k: h5py_kwargs[k] for k in h5py_kwargs & signature(h5py.File).parameters.keys()
    }
    h5py_kwargs = {k: h5py_kwargs[k] for k in h5py_kwargs - file_kwargs.keys()}
    if wo_mode == "write_safe":
        wo_mode = "w"
    if wo_mode == "append":
        wo_mode = "a"
    if wo_mode == "overwrite":
        wo_mode = "o"
    if wo_mode == "overwrite_file":
        wo_mode = "of"
        write_start = 0
    if wo_mode == "append_column":
        wo_mode = "ac"
    if wo_mode not in ["w", "a", "o", "of", "ac"]:
        msg = f"unknown wo_mode '{wo_mode}'"
        raise LH5EncodeError(msg, lh5_file, group, name)

    # "mode" is for the h5df.File and wo_mode is for this function
    # In hdf5, 'a' is really "modify" -- in addition to appending, you can
    # change any object in the file. So we use file:append for
    # write_object:overwrite.
    opened_here = False
    if not isinstance(lh5_file, h5py.File):
        mode = "w" if wo_mode == "of" or not Path(lh5_file).exists() else "a"

        try:
            fh = h5py.File(lh5_file, mode=mode, **file_kwargs)
        except OSError as oe:
            raise LH5EncodeError(oe, lh5_file) from oe

        opened_here = True
    else:
        fh = lh5_file

    try:
        log.debug(
            f"writing {obj!r}[{start_row}:{n_rows}] as "
            f"{fh.filename}:{group}/{name}[{write_start}:], "
            f"mode = {wo_mode}, h5py_kwargs = {h5py_kwargs}"
        )

        group = utils.get_h5_group(group, fh)

        # name already in file
        if name in group or (
            ("datatype" in group.attrs or group == "/")
            and (len(name) <= 2 or "/" not in name[1:-1])
        ):
            pass
        # group is in file but not struct or need to create nesting
        else:
            # check if name is nested
            # if name is nested, iterate up from parent
            # otherwise we just need to iterate the group
            if len(name) > 2 and "/" in name[1:-1]:
                group = utils.get_h5_group(
                    name[:-1].rsplit("/", 1)[0],
                    group,
                )
                curr_name = (
                    name.rsplit("/", 1)[1]
                    if name[-1] != "/"
                    else name[:-1].rsplit("/", 1)[1]
                )
            else:
                curr_name = name
            # initialize the object to be written
            obj = types.Struct({curr_name.replace("/", ""): obj})

            # if base group already has a child we just append
            if len(group) >= 1:
                wo_mode = "ac"
            else:
                # iterate up the group hierarchy until we reach the root or a group with more than one child
                while group.name != "/":
                    if len(group) > 1:
                        break
                    curr_name = group.name
                    group = group.parent
                    if group.name != "/":
                        obj = types.Struct({curr_name[len(group.name) + 1 :]: obj})
                    else:
                        obj = types.Struct({curr_name[1:]: obj})
                # if the group has more than one child, we need to append else we can overwrite
                wo_mode = "ac" if len(group) > 1 else "o"

            # set the new name
            if group.name == "/":
                name = "/"
            elif group.parent.name == "/":
                name = group.name[1:]
            else:
                name = group.name[len(group.parent.name) + 1 :]
            # get the new group
            group = utils.get_h5_group(group.parent if group.name != "/" else "/", fh)

        if wo_mode == "w" and name in group:
            msg = f"can't overwrite '{name}' in wo_mode 'write_safe'"
            raise LH5EncodeError(msg, fh, group, name)

        # struct, table, waveform table or histogram.
        if isinstance(obj, types.Struct):
            if (
                isinstance(obj, types.Histogram)
                and wo_mode not in ["w", "o", "of"]
                and name in group
            ):
                msg = f"can't append-write to histogram in wo_mode '{wo_mode}'"
                raise LH5EncodeError(msg, fh, group, name)
            if isinstance(obj, types.Histogram) and write_start != 0:
                msg = f"can't write histogram in wo_mode '{wo_mode}' with write_start != 0"
                raise LH5EncodeError(msg, fh, group, name)

            return _h5_write_struct(
                obj,
                name,
                fh,
                group=group,
                start_row=start_row,
                n_rows=n_rows,  # if isinstance(obj, types.Table | types.Histogram) else None,
                wo_mode=wo_mode,
                write_start=write_start,
                **h5py_kwargs,
            )

        # scalars
        if isinstance(obj, types.Scalar):
            return _h5_write_scalar(obj, name, fh, group, wo_mode)

        # vector of encoded vectors
        if isinstance(
            obj, (types.VectorOfEncodedVectors, types.ArrayOfEncodedEqualSizedArrays)
        ):
            group = utils.get_h5_group(
                name, group, grp_attrs=obj.attrs, overwrite=(wo_mode == "o")
            )

            # ask not to further compress flattened_data, it is already compressed!
            obj.encoded_data.flattened_data.attrs["compression"] = None

            _h5_write_vector_of_vectors(
                obj.encoded_data,
                "encoded_data",
                fh,
                group=group,
                start_row=start_row,
                n_rows=n_rows,
                wo_mode=wo_mode,
                write_start=write_start,
                **h5py_kwargs,
            )

            if isinstance(obj.decoded_size, types.Scalar):
                _h5_write_scalar(
                    obj.decoded_size,
                    "decoded_size",
                    fh,
                    group=group,
                    wo_mode=wo_mode,
                )
            else:
                _h5_write_array(
                    obj.decoded_size,
                    "decoded_size",
                    fh,
                    group=group,
                    start_row=start_row,
                    n_rows=n_rows,
                    wo_mode=wo_mode,
                    write_start=write_start,
                    **h5py_kwargs,
                )

            return None

        # vector of vectors
        if isinstance(obj, types.VectorOfVectors):
            return _h5_write_vector_of_vectors(
                obj,
                name,
                fh,
                group=group,
                start_row=start_row,
                n_rows=n_rows,
                wo_mode=wo_mode,
                write_start=write_start,
                **h5py_kwargs,
            )

        # if we get this far, must be one of the Array types
        if isinstance(obj, types.Array):
            return _h5_write_array(
                obj,
                name,
                fh,
                group=group,
                start_row=start_row,
                n_rows=n_rows,
                wo_mode=wo_mode,
                write_start=write_start,
                **h5py_kwargs,
            )

        msg = f"do not know how to write '{name}' of type '{type(obj).__name__}'"
        raise LH5EncodeError(msg, fh, group, name)
    finally:
        if opened_here:
            fh.close()


def _h5_write_struct(
    obj,
    name,
    lh5_file,
    group="/",
    start_row=0,
    n_rows=None,
    wo_mode="append",
    write_start=0,
    **h5py_kwargs,
):
    # this works for structs and derived (tables)
    assert isinstance(obj, types.Struct)

    # In order to append a column, we need to update the
    # `struct/table{old_fields}` value in `group.attrs['datatype"]` to include
    # the new fields. One way to do this is to override `obj.attrs["datatype"]`
    # to include old and new fields. Then we can write the fields to the
    # struct/table as normal.
    if wo_mode == "ac":
        if name not in group:
            msg = "Cannot append column to non-existing struct on disk"
            raise LH5EncodeError(msg, lh5_file, group, name)

        old_group = utils.get_h5_group(name, group)
        if "datatype" not in old_group.attrs:
            msg = "Cannot append column to an existing  non-LGDO object on disk"
            raise LH5EncodeError(msg, lh5_file, group, name)

        lgdotype = datatype.datatype(old_group.attrs["datatype"])
        fields = datatype.get_struct_fields(old_group.attrs["datatype"])
        if lgdotype is not type(obj):
            msg = (
                "Trying to append columns to an object of different "
                f"type {lgdotype.__name__}!={type(obj)}"
            )
            raise LH5EncodeError(msg, lh5_file, group, name)

        # If the mode is `append_column`, make sure we aren't appending
        # a table that has a column of the same name as in the existing
        # table. Also make sure that the field we are adding has the
        # same size
        if len(list(set(fields).intersection(set(obj.keys())))) != 0:
            msg = (
                f"Can't append {list(set(fields).intersection(set(obj.keys())))} "
                "column(s) to a table with the same field(s)"
            )
            raise LH5EncodeError(msg, lh5_file, group, name)

        # It doesn't matter what key we access, as all fields in the old table have the same size
        if (
            isinstance(obj, types.Table)
            and old_group.attrs["datatype"][:6]
            != "struct"  # structs dont care about size
            and old_group[next(iter(old_group.keys()))].size != obj.size
        ):
            msg = (
                f"Table sizes don't match. Trying to append column of size {obj.size} "
                f"to a table of size {old_group[next(iter(old_group.keys()))].size}."
            )
            raise LH5EncodeError(msg, lh5_file, group, name)

        # Now we can append the obj.keys() to the old fields, and then update obj.attrs.
        fields.extend(list(obj.keys()))
        obj.attrs.pop("datatype")

        obj.attrs["datatype"] = (
            obj.datatype_name() + "{" + ",".join(sorted(fields)) + "}"
        )

        # propagating wo_mode="ac" to nested LGDOs does not make any sense
        wo_mode = "append"

        # overwrite attributes of the existing struct
        attrs_overwrite = True
    else:
        attrs_overwrite = wo_mode == "o"

    group = utils.get_h5_group(
        name,
        group,
        grp_attrs=obj.attrs,
        overwrite=attrs_overwrite,
    )
    # If the mode is overwrite, then we need to peek into the file's
    # table's existing fields. If we are writing a new table to the
    # group that does not contain an old field, we should delete that
    # old field from the file
    if wo_mode == "o":
        # Find the old keys in the group that are not present in the
        # new table's keys, then delete them
        for key in list(set(group.keys()) - set(obj.keys())):
            log.debug(f"{key} is not present in new table, deleting field")
            del group[key]

    for field in obj:
        # eventually compress waveform table values with LGDO's
        # custom codecs before writing
        # if waveformtable.values.attrs["compression"] is NOT a
        # WaveformCodec, just leave it there
        obj_fld = None
        if (
            isinstance(obj, types.WaveformTable)
            and field == "values"
            and not isinstance(obj.values, types.VectorOfEncodedVectors)
            and not isinstance(obj.values, types.ArrayOfEncodedEqualSizedArrays)
            and "compression" in obj.values.attrs
            and isinstance(obj.values.attrs["compression"], compression.WaveformCodec)
        ):
            codec = obj.values.attrs["compression"]
            obj_fld = compression.encode(obj.values, codec=codec)
        else:
            obj_fld = obj[field]

        _h5_write_lgdo(
            obj_fld,
            str(field),
            lh5_file,
            group=group,
            start_row=start_row,
            n_rows=n_rows,
            wo_mode=wo_mode,
            write_start=write_start,
            **h5py_kwargs,
        )



from __future__ import annotations

import logging
from collections.abc import Callable, Collection, Iterator, Mapping
from concurrent.futures import Executor, ProcessPoolExecutor
from copy import deepcopy
from dataclasses import dataclass
from functools import partial
from itertools import chain, product
from typing import Any

import awkward as ak
import numpy as np
import pandas as pd
from hist import Hist, axis
from numpy.typing import NDArray

from ..types import LGDOCollection, Table
from ..units import default_units_registry as ureg
from .store import LH5Store
from .utils import expand_path


class LH5Iterator(Iterator):
    """Iterate over chunks of entries from LH5 files.

    The iterator reads ``buffer_len`` entries at a time from one or more
    files.  The LGDO instance returned at each iteration is reused to avoid
    reallocations, so copy the data if it should be preserved.

    Examples
    --------
    Iterate through a table one chunk at a time and call ``process`` on each chunk::

            from lgdo.lh5 import LH5Iterator
            for table in LH5Iterator("data.lh5", "geds/raw/energy", buffer_len=100):
                process(table)

    ``LH5Iterator`` can also be used for random access::

            it = LH5Iterator(files, groups)
            table = it.read(i_entry)

    In case of multiple files or an entry selection, ``i_entry`` refers to the
    global event index across all files.

    When instantiating an iterator you must provide a list of files and the
    HDF5 groups to read.  Optional parameters allow field masking, event
    selection and pairing the iterator with a "friend" iterator that is read in
    parallel.  Several properties are available to obtain the provenance of the
    data currently loaded:

    - ``current_i_entry`` -- index within the entry list of the first entry in
      the buffer
    - ``current_local_entries`` -- entry numbers relative to the file the data
      came from
    - ``current_global_entries`` -- entry number relative to the full dataset
    - ``current_files`` -- file name corresponding to each entry in the buffer
    - ``current_groups`` -- group name corresponding to each entry in the
      buffer
    """

    def __init__(
        self,
        lh5_files: str | Collection[str | Collection[str]],
        groups: str | Collection[str | Collection[str]],
        *,
        base_path: str = "",
        entry_list: Collection[int] | Collection[Collection[int]] = None,
        entry_mask: Collection[bool] | Collection[Collection[bool]] = None,
        i_start: int = 0,
        n_entries: int = None,
        field_mask: Mapping[str, bool] | Collection[str] = None,
        group_data: Mapping[Collection] | ak.Array = None,
        buffer_len: int = "100*MB",
        file_cache: int = 10,
        ds_map: NDArray[int] = None,
        friend: Collection[LH5Iterator] = None,
        friend_prefix: str = "",
        friend_suffix: str = "",
        h5py_open_mode: str = "r",
    ) -> None:
        """
        Constructor for LH5Iterator. Must provide a file or collection of
        files, and an lh5 group or collection of groups to read data from.

        Collections of files and groups can be nested. At the top level, we
        expect the same number of entries (one set of files to one set of groups).
        For each corresponding pair of sets, we will loop over each pairing of
        a file and group, with an inner loop over the groups to minimize the
        opening of files. Wildcards used for files will be expanded and applied
        in the inner loop (i.e. each file in a wildcard will read the same groups).
        If groups is an un-nested collection of strings, use all groups for all files.

        Examples
        --------
        Read "ch1/table" and "ch2/table" from "file.lh5"::

            LH5Iterator("/path/to/file.lh5", ["ch1/table", "ch2/table"])

        Read "ch1" from all lh5 files in "/path1", then read "ch1" and "ch2" from
        "/path2/file.lh5", and then read "ch1/2/3" from both "file1.lh5" and "file2.lh5"::

            LH5Iterator(
                ["/path1/*.lh5", "/path2/file.lh5", ["/path3/file1.lh5", "/path3/file2.lh5"]],
                ["ch1/table", ["ch1/table", "ch2/table"], ["ch1/table, "ch2/table, "ch3/table"]]
            )

        Parameters
        ----------
        lh5_files
            file(s) to read from (see above). May include wildcards and environment variables.
        groups
            HDF5 group(s) to read (see above).
        entry_list
            list of entry numbers to read. If a nested list is provided,
            expect one top-level list for each file, containing a list of
            local entries. If a list of ints is provided, use global entries.
        entry_mask
            mask of entries to read. If a list of arrays is provided, expect
            one for each file. Ignore if a selection list is provided.
        i_start
            index of first entry to start at when iterating
        n_entries
            number of entries to read before terminating iteration
        field_mask
            mask of which fields to read. See :meth:`LH5Store.read` for
            more details.
        group_data
            mapping of values corresponding to each provided lh5 group. Values
            will be duplicated for each entry in each dataset, corresponding to
            the correct group, and added to the output table. This should have
            same structure as ``groups``.
        buffer_len
            number of entries in tables yielded by iterator. Can be provided as
            a value with a unit of memory; in this case, use the estimated number
            of rows that will yield tables that require the provided memory.
            Default to 100*MB.
        file_cache
            maximum number of files to keep open at a time
        ds_map
            cumulative entries in datasets corresponding to file/group pairs.
            This can be provided on construction to speed up random or sparse
            access; otherwise, we sequentially read the size of each group.
            WARNING: no checks for accuracy are performed so only use this if
            you know what you are doing!
        friend
            a \"friend\" LH5Iterator that will be joined to this one, and read
            in parallel. The friend should have the same length and entry list.
            Each iteration will return a single LH5 Table containing columns from
            both iterators. The buffer_len will be set to the minimum of the two.
        friend_prefix
            prefix for fields in friend iterator for resolving naming conflicts
        friend_suffix
            suffix for fields in friend iterator for resolving naming conflicts
        h5py_open_mode
            file open mode used when acquiring file handles. ``r`` (default)
            opens files read-only while ``a`` allow opening files for
            write-appending as well.
        """

        if h5py_open_mode == "read":
            h5py_open_mode = "r"
        if h5py_open_mode == "append":
            h5py_open_mode = "a"
        if h5py_open_mode not in ["r", "a"]:
            msg = f"unknown h5py_open_mode '{h5py_open_mode}'"
            raise ValueError(msg)

        self.lh5_st = LH5Store(
            base_path=base_path, keep_open=file_cache, default_mode=h5py_open_mode
        )

        # convert lh5_files into a nested list
        if isinstance(lh5_files, str):
            self.lh5_files = [expand_path(lh5_files, list=True, base_path=base_path)]
        elif not isinstance(lh5_files, Collection):
            msg = "lh5_files must be a string or list of strings"
            raise ValueError(msg)
        else:
            self.lh5_files = []
            for f in lh5_files:
                if isinstance(f, str):
                    self.lh5_files.append(
                        expand_path(f, list=True, base_path=base_path)
                    )
                elif isinstance(f, Collection) and all(
                    isinstance(name, str) for name in f
                ):
                    self.lh5_files.append(
                        [
                            path
                            for name in f
                            for path in expand_path(
                                name, list=True, base_path=base_path
                            )
                        ]
                    )
                else:
                    msg = "lh5_files must be a collection of strings with up to two levels of nesting"

        # convert groups into a nested list
        if isinstance(groups, str):
            self.groups = [[groups]] * len(self.lh5_files)
        elif not isinstance(groups, Collection):
            msg = "group must be a string or collection of strings"
            raise ValueError(msg)
        elif all(isinstance(g, str) for g in groups):
            self.groups = [groups] * len(self.lh5_files)
        else:
            self.groups = []
            for g in groups:
                if isinstance(g, str):
                    g = [g]  # noqa: PLW2901
                elif not isinstance(g, Collection) and all(
                    isinstance(name, str) for name in g
                ):
                    msg = "groups must be a collection of strings with up to two levels of nesting"
                    raise ValueError(msg)
                self.groups.append(g)

        # convert group_data into array of records
        if isinstance(group_data, pd.DataFrame):
            self.group_data = ak.Array(dict(group_data))
        elif group_data is not None:
            self.group_data = ak.Array(group_data)
        else:
            self.group_data = None

        if isinstance(self.group_data, ak.Array):
            if not self.group_data.fields:
                msg = "group_data must have named fields"
                raise ValueError(msg)
            self.group_data = ak.zip(
                {f: self.group_data[f] for f in self.group_data.fields}
            )
            if self.group_data.ndim == 1:
                self.group_data = ak.Array([self.group_data] * len(self.lh5_files))
            if not all(
                len(gd) == len(gps)
                for gd, gps in zip(self.group_data, self.groups, strict=False)
            ):
                msg = "group_data must have same array structure as groups"
                raise ValueError(msg)

        # check that groups and lh5_files are compatible
        if len(self.groups) != len(self.lh5_files):
            msg = "lh5_files and groups must have same length at first level of nesting"
            raise ValueError(msg)

        # outer-product and flatten nested lists to get all group/file combinations for datasets
        file_list = []
        group_list = []
        group_data_list = [] if self.group_data is not None else None
        for i in range(len(self.lh5_files)):
            f_list = self.lh5_files[i]
            g_list = self.groups[i]
            gd_list = self.group_data[i] if self.group_data is not None else None
            for i_f, i_g in product(range(len(f_list)), range(len(g_list))):
                file_list.append(f_list[i_f])
                group_list.append(g_list[i_g])
                if gd_list is not None:
                    group_data_list.append(gd_list[i_g])
        self.lh5_files = file_list
        self.groups = group_list
        self.group_data = group_data_list
        self.n_datasets = len(self.lh5_files)

        if entry_list is not None and entry_mask is not None:
            msg = "entry_list and entry_mask arguments are mutually exclusive"
            raise ValueError(msg)

        # Map to last row in each file
        if ds_map is None:
            self.ds_map = np.full(self.n_datasets, np.iinfo("q").max, "q")
        else:
            self.ds_map = np.array(ds_map)

        # Map to last iterator entry for each file
        self.entry_map = np.full(self.n_datasets, np.iinfo("q").max, "q")

        self.friend = []
        self.friend_prefix = []
        self.friend_suffix = []

        if self.n_datasets == 0:
            msg = f"can't open any files from {lh5_files}"
            raise RuntimeError(msg)

        # lh5 buffer will contain all fields and be used for I/O (with a field mask)
        self.lh5_buffer = self.lh5_st.get_buffer(
            self.groups[0],
            self.lh5_files[0],
            size=0,
        )

        def get_available_fields(tab):
            ret = set()
            if not isinstance(tab, Table):
                return ret

            for k, v in tab.items():
                ret |= {k}
                if isinstance(v, Table):
                    ret |= {f"{k}/{field}" for field in get_available_fields(v)}
            return ret

        self.available_fields = get_available_fields(self.lh5_buffer)

        # set field mask and buffer length
        # Note: group_data is added to table here!
        self.reset_field_mask(field_mask)
        self.buffer_len = buffer_len

        # Attach the friend(s)
        if friend is None:
            friend = []
        elif isinstance(friend, LH5Iterator):
            friend = [friend]

        if len(friend) > 0:
            fr_buf_len = min(fr.buffer_len for fr in friend)
            self.buffer_len = min(self.buffer_len, fr_buf_len)

        if isinstance(friend_prefix, str):
            friend_prefix = [friend_prefix] * len(friend)
        if isinstance(friend_suffix, str):
            friend_suffix = [friend_suffix] * len(friend)
        for fr, prefix, suffix in zip(
            friend, friend_prefix, friend_suffix, strict=False
        ):
            self.add_friend(fr, prefix, suffix)

        self.i_start = i_start
        self.n_entries = n_entries
        self.current_i_entry = 0
        self.next_i_entry = 0

        # List of entry indices from each file
        self.local_entry_list = None
        self.global_entry_list = None
        if entry_list is not None:
            entry_list = list(entry_list)
            if isinstance(entry_list[0], int):
                self.local_entry_list = [None] * self.n_datasets
                self.global_entry_list = np.array(entry_list, "q")
                self.global_entry_list.sort()

            else:
                self.local_entry_list = [[]] * self.n_datasets
                for i_ds, local_list in enumerate(entry_list):
                    self.local_entry_list[i_ds] = np.array(local_list, "q")
                    self.local_entry_list[i_ds].sort()

        elif entry_mask is not None:
            # Convert entry mask into an entry list
            if isinstance(entry_mask, pd.Series):
                entry_mask = entry_mask.to_numpy()
            if isinstance(entry_mask, np.ndarray):
                self.local_entry_list = [None] * self.n_datasets
                self.global_entry_list = np.nonzero(entry_mask)[0]
            else:
                self.local_entry_list = [[]] * self.n_datasets
                for i_ds, local_mask in enumerate(entry_mask):
                    self.local_entry_list[i_ds] = np.nonzero(local_mask)[0]

    def _get_ds_cumlen(self, i_ds: int) -> int:
        """Helper to get cumulative dataset length of file/groups"""
        if i_ds < 0:
            return 0
        fcl = self.ds_map[i_ds]

        # if we haven't already calculated, calculate for all files up to i_ds
        if fcl == np.iinfo("q").max:
            i_start = np.searchsorted(self.ds_map, np.iinfo("q").max)
            fcl = self.ds_map[i_start - 1] if i_start > 0 else 0

            for i in range(i_start, i_ds + 1):
                fcl += self.lh5_st.read_n_rows(self.groups[i], self.lh5_files[i])
                self.ds_map[i] = fcl
        return fcl

    def _get_ds_cumentries(self, i_ds: int) -> int:
        """Helper to get cumulative iterator entries in file/groups"""
        if i_ds < 0:
            return 0
        n = self.entry_map[i_ds]

        # if we haven't already calculated, calculate for all files up to i_ds
        if n == np.iinfo("q").max:
            i_start = np.searchsorted(self.entry_map, np.iinfo("q").max)
            n = self.entry_map[i_start - 1] if i_start > 0 else 0

            for i in range(i_start, i_ds + 1):
                elist = self.get_ds_entrylist(i)
                fcl = self._get_ds_cumlen(i)
                if elist is None:
                    # no entry list provided
                    n = fcl
                else:
                    n += len(elist)
                    # check that file entries fall inside of file
                    if len(elist) > 0 and elist[-1] >= fcl:
                        logging.warning(f"Found entries out of range for file {i}")
                        n += np.searchsorted(elist, fcl, "right") - len(elist)
                self.entry_map[i] = n
        return n

    def get_ds_entrylist(self, i_ds: int) -> np.ndarray:
        """Helper to get entry list for dataset"""
        # If no entry list is provided
        if self.local_entry_list is None:
            return None

        elist = self.local_entry_list[i_ds]
        if elist is None:
            # Get local entrylist for this dataset from global entry list
            f_start = self._get_ds_cumlen(i_ds - 1)
            f_end = self._get_ds_cumlen(i_ds)
            i_start = self._get_ds_cumentries(i_ds - 1)
            i_stop = np.searchsorted(self.global_entry_list, f_end, "right")
            elist = np.array(self.global_entry_list[i_start:i_stop], "q") - f_start
            self.local_entry_list[i_ds] = elist
        return elist

    def get_global_entrylist(self) -> np.ndarray:
        """Get global entry list, constructing it if needed"""
        if self.global_entry_list is None and self.local_entry_list is not None:
            self.global_entry_list = np.zeros(len(self), "q")
            for i_ds in range(self.n_datasets):
                i_start = self._get_ds_cumentries(i_ds - 1)
                i_stop = self._get_ds_cumentries(i_ds)
                f_start = self._get_ds_cumlen(i_ds - 1)
                self.global_entry_list[i_start:i_stop] = (
                    self.get_ds_entrylist(i_ds) + f_start
                )
        return self.global_entry_list

    def read(self, i_entry: int, n_entries: int | None = None) -> Table:
        "Read the nextlocal chunk of events, starting at entry."
        self.lh5_buffer.resize(0)

        if n_entries is None:
            n_entries = self.buffer_len
        elif n_entries == 0:
            return self.lh5_buffer
        elif n_entries > self.buffer_len:
            msg = "n_entries cannot be larger than buffer_len"
            raise ValueError(msg)

        # if dataset hasn't been opened yet, search through datasets
        # sequentially until we find the right one
        i_ds = np.searchsorted(self.entry_map, i_entry, "right")
        if i_ds < self.n_datasets and self.entry_map[i_ds] == np.iinfo("q").max:
            while i_ds < self.n_datasets and i_entry >= self._get_ds_cumentries(i_ds):
                i_ds += 1

        if i_ds == self.n_datasets:
            return self.lh5_buffer
        local_i_entry = i_entry - self._get_ds_cumentries(i_ds - 1)

        while len(self.lh5_buffer) < n_entries and i_ds < self.n_datasets:
            # Loop through datasets
            local_idx = self.get_ds_entrylist(i_ds)
            if local_idx is not None and len(local_idx) == 0:
                i_ds += 1
                local_i_entry = 0
                continue

            i_local = local_i_entry if local_idx is None else local_idx[local_i_entry]
            self.lh5_buffer = self.lh5_st.read(
                self.groups[i_ds],
                self.lh5_files[i_ds],
                start_row=i_local,
                n_rows=n_entries - len(self.lh5_buffer),
                idx=local_idx,
                field_mask=self.field_mask,
                obj_buf=self.lh5_buffer,
                obj_buf_start=len(self.lh5_buffer),
            )

            if self.group_data is not None:
                data = self.group_data[i_ds]
                for f in data.fields:
                    self.lh5_buffer[f][local_idx:] = data[f]

            i_ds += 1
            local_i_entry = 0

        self.current_i_entry = i_entry

        for friend in self.friend:
            friend.read(i_entry)

        return self.lh5_buffer

    @property
    def buffer_len(self):
        return self._buffer_len

    @buffer_len.setter
    def buffer_len(self, buffer_len: str | ureg.Quantity | int):
        if isinstance(buffer_len, str):
            buffer_len = ureg.Quantity(buffer_len)
        if isinstance(buffer_len, ureg.Quantity):
            f = self.lh5_files[0]
            g = self.groups[0]
            buffer_len = int(
                buffer_len
                / (self.lh5_st.read_size_in_bytes(g, f) * ureg.B)
                * self.lh5_st.read_n_rows(g, f)
            )

        self._buffer_len = buffer_len
        for fr in self.friend:
            fr.buffer_len = buffer_len

    def add_friend(self, friend: LH5Iterator, prefix: str = "", suffix: str = ""):
        """Add a friend which will be iterated alongside this, returning a Table
        joining the contents of each.

        Parameters
        ----------
        friend
            LH5Iterator to be friended to this one
        prefix
            string prepended to field names; useful for disambiguating conflicts
        suffix
            string appended to field names; useful for disambiguating conflicts
        """
        if not isinstance(friend, LH5Iterator):
            msg = "Friend must be an LH5Iterator"
            raise ValueError(msg)

        # set buffer_lens to be equal
        if friend.buffer_len > self.buffer_len:
            friend.buffer_len = self.buffer_len
        elif friend.buffer_len < self.buffer_len:
            self.buffer_len = friend.buffer_len
        friend.lh5_buffer.resize(len(self.lh5_buffer))

        self.friend += [friend]
        self.friend_prefix += [prefix]
        self.friend_suffix += [suffix]
        self.lh5_buffer.join(
            friend.lh5_buffer,
            keep_mine=True,
            prefix=prefix,
            suffix=suffix,
        )

    def reset_field_mask(
        self,
        mask: Collection[str]
        | Mapping[str, bool]
        | Collection[Collection[str]]
        | Collection[Mapping[str, bool]]
        | None,
        warn_missing=True,
    ):
        """Replaces the field mask of this iterator and any friends with mask.

        - If ``None``, set this and all friends to have no mask.
        - If a collection of strings or mapping from strings to bools, set the mask
          for this and all friends; in the case of a conflict, use first column found. If a
          prefix or suffix is included for the friend, it must be included in this mask
        - If a collection of collections, use the first item to set this mask, and subsequent
          items to set friend masks. In this case, do not include prefixes or suffixes in names
        """
        if mask is None:
            self.field_mask = self.available_fields

            for fr in self.friend:
                fr.reset_field_mask(None)

            remaining_fields = []

        elif isinstance(mask, Mapping):
            self.field_mask = {
                field: mask[field] for field in self.available_fields if field in mask
            }
            mask = {
                field: mask[field] for field in mask if field not in self.field_mask
            }

            for fr, pre, suf in zip(
                self.friend, self.friend_prefix, self.friend_suffix, strict=False
            ):
                mask_lookup = {
                    f"{pre}{field}{suf}": field for field in fr.available_fields
                }
                fr_mask = {
                    mask_lookup[field]: mask[field]
                    for field in mask_lookup
                    if field in mask
                }
                fr.reset_field_mask(fr_mask)
                mask = {
                    field: mask[field] for field in mask if field not in mask_lookup
                }

            remaining_fields = mask

        elif isinstance(mask, Collection) and all(isinstance(m, str) for m in mask):
            mask = set(mask)
            self.field_mask = mask & set(self.available_fields)
            mask -= self.field_mask

            for fr, pre, suf in zip(
                self.friend, self.friend_prefix, self.friend_suffix, strict=False
            ):
                mask_lookup = {
                    f"{pre}{field}{suf}": field for field in fr.available_fields
                }
                fr_mask = {mask_lookup[field] for field in mask if field in mask_lookup}
                fr.reset_field_mask(fr_mask)
                mask -= set(mask_lookup)

            remaining_fields = mask

        # Create a new buffer and move any elements from the old into the new
        def copy_data(old_buffer, new_buffer):
            if isinstance(new_buffer, Table):
                for k, v in new_buffer.items():
                    if k in old_buffer:
                        new_buffer[k] = copy_data(v, old_buffer[k])
                return new_buffer
            return old_buffer

        self.lh5_buffer = copy_data(
            self.lh5_buffer,
            self.lh5_st.get_buffer(
                self.groups[0],
                self.lh5_files[0],
                size=len(self.lh5_buffer),
                field_mask=self.field_mask,
            ),
        )

        for fr, pre, suf in zip(
            self.friend, self.friend_prefix, self.friend_suffix, strict=False
        ):
            self.lh5_buffer.join(
                fr.lh5_buffer,
                keep_mine=True,
                prefix=pre,
                suffix=suf,
            )

        # join Table with group metadata; repeat first record to initialize
        if self.group_data:
            tb_gd = deepcopy(Table(ak.Array([self.group_data[0]])))
            tb_gd.resize(len(self.lh5_buffer))
            self.lh5_buffer.join(tb_gd)

        if warn_missing and len(remaining_fields) > 0:
            logging.warning(f"Fields {remaining_fields} in field mask were not found")

    @property
    def current_local_entries(self) -> NDArray[int]:
        """Return list of local dataset entries in buffer"""
        cur_entries = np.zeros(len(self.lh5_buffer), dtype="int32")
        i_ds = np.searchsorted(self.entry_map, self.current_i_entry, "right")
        ds_start = self._get_ds_cumentries(i_ds - 1)
        i_local = self.current_i_entry - ds_start
        i = 0

        while i < len(cur_entries):
            # number of entries to read from this file
            ds_end = self._get_ds_cumentries(i_ds)
            n = min(ds_end - ds_start - i_local, len(cur_entries) - i)
            entries = self.get_ds_entrylist(i_ds)

            if entries is None:
                cur_entries[i : i + n] = np.arange(i_local, i_local + n)
            else:
                cur_entries[i : i + n] = entries[i_local : i_local + n]

            i_ds += 1
            ds_start = ds_end
            i_local = 0
            i += n

        return cur_entries

    @property
    def current_global_entries(self) -> NDArray[int]:
        """Return list of local file entries in buffer"""
        cur_entries = np.zeros(len(self.lh5_buffer), dtype="int32")
        i_ds = np.searchsorted(self.entry_map, self.current_i_entry, "right")
        ds_start = self._get_ds_cumentries(i_ds - 1)
        i_local = self.current_i_entry - ds_start
        i = 0

        while i < len(cur_entries):
            # number of entries to read from this file
            ds_end = self._get_ds_cumentries(i_ds)
            n = min(ds_end - ds_start - i_local, len(cur_entries) - i)
            entries = self.get_ds_entrylist(i_ds)

            if entries is None:
                cur_entries[i : i + n] = self._get_ds_cumlen(i_ds - 1) + np.arange(
                    i_local, i_local + n
                )
            else:
                cur_entries[i : i + n] = (
                    self._get_ds_cumlen(i_ds - 1) + entries[i_local : i_local + n]
                )

            i_ds += 1
            ds_start = ds_end
            i_local = 0
            i += n

        return cur_entries

    @property
    def current_files(self) -> NDArray[str]:
        """Return list of file names for entries in buffer"""
        cur_files = np.zeros(len(self.lh5_buffer), dtype=np.dtypes.StringDType)
        i_ds = np.searchsorted(self.entry_map, self.current_i_entry, "right")
        ds_start = self._get_ds_cumentries(i_ds - 1)
        i_local = self.current_i_entry - ds_start
        i = 0

        while i < len(cur_files):
            # number of entries to read from this file
            ds_end = self._get_ds_cumentries(i_ds)
            n = min(ds_end - ds_start - i_local, len(cur_files) - i)
            cur_files[i : i + n] = self.lh5_files[i_ds]

            i_ds += 1
            ds_start = ds_end
            i_local = 0
            i += n

        return cur_files

    @property
    def current_groups(self) -> NDArray[str]:
        """Return list of group names for entries in buffer"""
        cur_groups = np.zeros(len(self.lh5_buffer), dtype=np.dtypes.StringDType)
        i_ds = np.searchsorted(self.entry_map, self.current_i_entry, "right")
        ds_start = self._get_ds_cumentries(i_ds - 1)
        i_local = self.current_i_entry - ds_start
        i = 0

        while i < len(cur_groups):
            # number of entries to read from this file
            ds_end = self._get_ds_cumentries(i_ds)
            n = min(ds_end - ds_start - i_local, len(cur_groups) - i)
            cur_groups[i : i + n] = self.groups[i_ds]

            i_ds += 1
            ds_start = ds_end
            i_local = 0
            i += n

        return cur_groups

    def __len__(self) -> int:
        """Return the total number of entries to be read."""
        if len(self.entry_map) == 0:
            return 0
        if self.n_entries is None:
            return self._get_ds_cumentries(self.n_datasets - 1)
        # only check as many files as we strictly need to
        for i in range(self.n_datasets):
            if self.n_entries < self._get_ds_cumentries(i):
                return self.n_entries
        return self._get_ds_cumentries(self.n_datasets - 1)

    def __iter__(self) -> LH5Iterator:
        """Loop through entries in blocks of size buffer_len."""
        self.current_i_entry = 0
        self.next_i_entry = self.i_start
        return self

    def __next__(self) -> Table:
        """Read next buffer_len entries and return lh5_table and iterator entry."""
        n_entries = self.n_entries
        if n_entries is not None:
            n_entries = min(
                self.buffer_len, n_entries + self.i_start - self.next_i_entry
            )

        buf = self.read(self.next_i_entry, n_entries)
        if len(buf) == 0:
            raise StopIteration
        self.next_i_entry = self.current_i_entry + len(buf)
        return buf

    def __deepcopy__(self, memo):
        """Deep copy everything except lh5_st"""
        result = LH5Iterator.__new__(LH5Iterator)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            if k == "lh5_st":
                result.lh5_st = LH5Store(
                    base_path=self.lh5_st.base_path, keep_open=self.lh5_st.keep_open
                )
            else:
                setattr(result, k, deepcopy(v, memo))
        return result

    def __getstate__(self):
        """Do not try to pickle lh5_st or lh5_buffer"""
        return dict(
            self.__dict__,
            lh5_st={
                "base_path": self.lh5_st.base_path,
                "keep_open": self.lh5_st.keep_open,
            },
            lh5_buffer=None,
        )

    def __setstate__(self, d):
        """Reinitialize lh5_st and lh5_buffer to avoid potential issues"""
        self.__dict__ = d
        self.lh5_st = LH5Store(**(d["lh5_st"]))
        self.lh5_st.gimme_file(self.lh5_files[0])
        self.lh5_buffer = self.lh5_st.get_buffer(
            self.groups[0],
            self.lh5_files[0],
            size=self.buffer_len,
            field_mask=self.field_mask,
        )
        for fr, pre, suf in zip(
            self.friend, self.friend_prefix, self.friend_suffix, strict=False
        ):
            self.lh5_buffer.join(
                fr.lh5_buffer,
                keep_mine=True,
                prefix=pre,
                suffix=suf,
            )

    def _select_groups(self, i_beg, i_end):
        """Reduce list of files and groups; used by _generate_workers"""
        s = slice(i_beg, i_end)
        self.lh5_files = self.lh5_files[s]
        self.groups = self.groups[s]
        self.n_datasets = i_end - i_beg

        if i_beg > 0:
            np.subtract(
                self.ds_map,
                self.ds_map[i_beg - 1],
                out=self.ds_map,
                where=self.ds_map != np.iinfo("q").max,
            )
            np.subtract(
                self.entry_map,
                self.entry_map[i_beg - 1],
                out=self.entry_map,
                where=self.entry_map != np.iinfo("q").max,
            )
        self.ds_map = self.ds_map[s]
        self.entry_map = self.entry_map[s]

        if self.local_entry_list is not None:
            self.local_entry_list = self.local_entry_list[s]
        self.global_entry_list = None

        for fr in self.friend:
            fr._select_groups(i_beg, i_end)

    def _generate_workers(self, n_workers: int):
        """Create n_workers copy of this iterator, dividing the datasets (file/groups)
        groups between them. These are intended for parallel use"""
        i_datasets = np.linspace(0, len(self.lh5_files), n_workers + 1).astype("int")
        # if we have an entry list, get local entries for all files
        if self.local_entry_list is not None:
            for i in range(len(self.lh5_files)):
                self.get_ds_entrylist(i)

        worker_its = []
        for i_worker in range(n_workers):
            it = deepcopy(self)
            it._select_groups(i_datasets[i_worker], i_datasets[i_worker + 1])
            worker_its += [it]

        return worker_its

    def map(
        self,
        fun: Callable[Table, LH5Iterator, Any],
        aggregate: Callable = None,
        init: Any = None,
        begin: Callable[LH5Iterator] = None,
        terminate: Callable[LH5Iterator] = None,
        processes: int = None,
        executor: Executor = None,
    ) -> Iterator[Any]:
        """Map function over iterator blocks.

        Returns order-preserving list of outputs. Can be multi-threaded
        provided there are no attempts to modify existing objects.
        Multi-threading splits the iterator into multiple independent
        streams with an approximately equal number of files/groups,
        concurrently processed under a single program multiple data
        model. Results will be returned asynchronously for each process.

        Note: see :meth:`query` and :meth:`hist`

        Example
        -------
        Process a table and sum the products at the end::

            def process(lh5_tab, lh5_it):
                ...process the table
                return result_of_processing

            results = lh5_it.map(process, processes=4)
            # results are an iterator over lists
            result = sum(val for result in results for val in result)

        Process a table as above, using aggregate to sum the results::

            def process(lh5_tab, lh5_it):
                ...process the table
                return result_of_processing

            result = lh5_it.map(process, aggregate=np.add, init=0, processes=4)

        Process a table using a more arbitrary output::

            class Result:
                def __init__(self):
                    ...initialize

                @classmethod
                def process_table(tab):
                    ...process the table

                def aggregate(self, result):
                    ...add data from processing into object

            result = lh5_it.map(
                Result.process_table,
                aggregate=Result.aggregate,
                init=Result(),
                processes=4
            )

        Parameters
        ----------
        fun:
            function with signature ``fun(lh5_obj: Table, it: LH5Iterator) -> Any``
            Outputs of function will be collected in list and returned
        aggregate:
            function used to iterably combine outputs of ``fun`` for each block
            of data. Should have two inputs; first input should be the type of
            the aggregate, and second of the type returned by ``fun``. This function
            can either return the result, or perform the aggregation in-place on
            the first element and return ``None``. If using multi-processing, ``map``
            will return an async-iterator over the aggregated results from each process.
            If ``None``, do not aggregate and instead return will iterate over
            result for each block.
        init:
            initial value used for aggregation. If using an aggregating function
            and ``init`` is ``None``, perform a deep copy of the first element
        begin:
            function with signature ``fun(it: LH5Iterator)`` that is run before we
            loop through a chunk of the iterator
        terminate:
            function with the signature ``fun(it: LH5Iterator)`` that is run after
            we finish looping through a chunk of the iterator
        processes:
            number of processes. If ``None``, use number equal to threads available
            to ``executor`` (if provided), or else do not parallelize
        executor:
            :class:`concurrent.futures.Executor` object for managing parallelism.
            If ``None``, create a :class:`concurrent.futures.`ProcessPoolExecutor`
            with number of processes equal to ``processes``.
        """

        # if no aggregate is provided, append results to a list
        if aggregate is None:
            init = []
            aggregate = _append_copy

        if processes is None:
            if isinstance(executor, Executor):
                processes = executor._max_workers
            else:
                return _map_helper(fun, aggregate, init, begin, terminate, self)

        if executor is None:
            executor = ProcessPoolExecutor(processes)

        it_pool = self._generate_workers(processes)

        result = executor.map(
            partial(_map_helper, fun, aggregate, init, begin, terminate), it_pool
        )

        # If no aggregator was given, chain iterators
        if aggregate is _append_copy:
            return chain.from_iterable(result)
        return result

    def query(
        self,
        where: Callable | str,
        processes: Executor | int = None,
        executor: Executor = None,
        library: str = None,
    ):
        """
        Query the data files in the iterator

        Returns the selected data as a single table in one of several formats.

        Examples
        --------
        Query data using a string selection::

            tab = lh5_it.query("(col1 == 0) & (col2 > 100)")

        Query data using a function::

            def select(lh5_tab, lh5_it):
                ...process data and produce a new table
                return result

            tab = lh5_it.query(select)

        Parameters
        ----------
        where:
            A filter function for selecting data entries. Can be:

            - A function that returns reduced data, with signature
              ``fun(lh5_obj: Table, it: LH5Iterator)``. Can return:

              - :class:`numpy.ndarray`: if 1D list of values; if 2D list of lists of
                values in same order as axes
              - ``Collection[ArrayLike]``: return list of values in same order as axes
              - ``Mapping[str, ArrayLike]``: mapping from axis name to values
              - ``pandas.DataFrame``: pandas dataframe. Treat as mapping from column
                name to values

            - A string expression. This will call `Table.eval` to select events and
                return a table with all fields in the `field_mask`, in a data format
                provided with `library`.

        processes:
            number of processes. If ``None``, use number equal to threads available
            to ``executor`` (if provided), or else do not parallelize
        executor:
            :class:`concurrent.futures.Executor` object for managing parallelism.
            If ``None``, create a :class:`concurrent.futures.`ProcessPoolExecutor`
            with number of processes equal to ``processes``.
        library:
            library to convert the columns to when using a string expression for ``where``.
            See :meth:`Table.eval`.
        """
        if where is None:
            where = _identity

        if isinstance(where, str):
            where = _table_query(where, library)

        test = where(self.lh5_buffer, self)
        if isinstance(test, LGDOCollection):
            it = self.map(
                where, processes=processes, executor=executor, aggregate=Table.append
            )
            if isinstance(it, LGDOCollection):
                return it
            ret = next(it)
            for res in it:
                ret.append(res)
            return ret
        if isinstance(test, pd.DataFrame):
            return pd.concat(
                list(self.map(where, processes=processes, executor=executor)),
                ignore_index=True,
            )
        if isinstance(test, np.ndarray):
            return np.concatenate(
                list(self.map(where, processes=processes, executor=executor))
            )
        if isinstance(test, ak.Array):
            return ak.concatenate(
                list(self.map(where, processes=processes, executor=executor))
            )

        msg = f"Cannot call query with return type {test.__class__}. "
        "Allowed return types: LGDOCollection, np.array, pd.DataFrame, ak.Array"
        raise ValueError(msg)

    def hist(
        self,
        ax: Hist | axis | Collection[axis],
        where: Callable | str = None,
        keys: Collection[str] | str = None,
        processes: Executor | int = None,
        executor: Executor = None,
        **hist_kwargs,
    ) -> Hist:
        """
        Fill a histogram from data produced by a query selecting on ``where``. If
        ``where`` is ``None``, fill with all data fetched by iterator.

        Examples
        --------
        Build a 1D histogram of values in ``col3`` with a string selection::

            h = lh5_it.query(
                hist.axis.RegularAxis(100, 0, 500, label="col3")
                where = "(col1 == 0) & (col2 > 100)",
                keys = "col3"
            )

        Build a 2D histogram with a value axis and string-category axis after
        applying some processing::

            def get_val(lh5_tab, lh5_it):
                ...process data
                return value, category

            h = lh5_it
                [hist.axis.RegularAxis(100, 0, 500, label="Value"),
                 hist.axis.StrCategory([], growth=True, label="Category")],
                 where = get_val,
            )

        Parameters
        ----------
        ax:
            :class:`hist.axis` object(s) used to construct the histogram. Can provide a
            :class:`hist.Hist` which will be filled as well.
        where:
            A filter function for selecting data entries to put into the histogram. Can be:

            - A function that returns reduced data, with signature
              ``fun(lh5_obj: Table, it: LH5Iterator)``. Can return:

              - :class:`numpy.ndarray`: if 1D list of values; if 2D list of lists of
                values in same order as axes
              - ``Collection[ArrayLike]``: return list of values in same order as axes
              - ``Mapping[str, ArrayLike]``: mapping from axis name to values
              - :class:`pandas.DataFrame`: treat as mapping from column name to values
            - A string expression. This will call :meth:`lgdo.Table.eval` and return
              a pandas DataFrame containing all columns in the fields mask.

        keys:
            list of keys fields corresponding to axes. Use if where
            returns a mapping with names different from axis names.
        processes:
            number of processes. If ``None``, use number equal to threads available
            to ``executor`` (if provided), or else do not parallelize
        executor:
            :class:`concurrent.futures.Executor` object for managing parallelism.
            If ``None``, create a :class:`concurrent.futures.`ProcessPoolExecutor`
            with number of processes equal to ``processes``.
        hist_kwargs:
            additional keyword arguments for constructing Hist. See :class:`hist.Hist`.
        """

        # get initial hist for each thread
        if isinstance(ax, axis.AxesMixin):
            h = Hist(ax, **hist_kwargs)
        elif isinstance(ax, Collection):
            h = Hist(*ax, **hist_kwargs)
        elif isinstance(ax, Hist):
            h = ax.copy()
            h[...] = 0

        if where is None:
            where = _identity
        elif isinstance(where, str):
            where = _table_query(where, "ak")

        h = self.map(
            where,
            processes=processes,
            executor=executor,
            aggregate=_hist_filler(keys),
            init=h,
        )
        if isinstance(h, Iterator):
            h = sum(h)

        if isinstance(ax, Hist):
            ax += h
            return ax
        return h


# Would that python multiprocessing allowed lambdas...
def _identity(val, _):
    return val


def _append_copy(list, val):
    "Helper for aggregating tables in query"
    list.append(deepcopy(val))


def _map_helper(fun, aggregator, init, begin, terminate, it):
    "Helper for executing int, begin and terminate functions when calling map"
    if begin:
        begin(it)

    aggregate = init
    for tab in it:
        result = fun(tab, it)

        if aggregate is None:
            # if no init, initialize on first entry
            aggregate = deepcopy(result)
        else:
            res = aggregator(aggregate, result)
            if res is not None:
                aggregate = res

    if terminate:
        terminate(it)

    return aggregate


@dataclass
class _table_query:
    "Helper for when query is called on a string"

    expr: str
    library: str

    def __call__(self, tab, _):
        "Evaluate selection and return selected elements"
        # Note this could probably be done more simply if we
        # implemented __getitem__ to work with lists/slices
        mask = tab.eval(self.expr)
        ret = tab.view_as("ak")[mask]
        if self.library == "ak":
            return ret
        if self.library is None:
            return Table(ret)
        return Table(ret).view_as(self.library)


class _hist_filler:
    "Helper for filling histogram"

    def __init__(self, keys):
        if keys is not None:
            if isinstance(keys, str):
                keys = [keys]
            elif not isinstance(keys, list):
                keys = list(keys)
        self.keys = keys

    def __call__(self, hist, data):
        if isinstance(data, np.ndarray) and len(data.shape) == 1:
            hist.fill(data)
        elif isinstance(data, np.ndarray) and len(data.shape) == 2:
            hist.fill(*data)
        elif isinstance(data, (pd.DataFrame, Mapping)):
            if self.keys is not None:
                hist.fill(*[data[k] for k in self.keys])
            else:
                hist.fill(**data)
        elif isinstance(data, ak.Array):
            # TODO: handle nested ak arrays
            if self.keys is not None:
                hist.fill(*[data[k] for k in self.keys])
            else:
                hist.fill(*[data[f] for f in data.fields])
        elif isinstance(data, Collection):
            hist.fill(*data)
        else:
            msg = "data returned by where is not compatible with hist. Must be a 1d or 2d numpy array, a list of arrays, or a mapping from str to array"
            raise ValueError(msg)



from __future__ import annotations

import pickle

import awkward as ak
import numpy as np
import pandas as pd
import pint
import pytest

from lgdo import Array


def test_datatype_name():
    array = Array()
    assert array.datatype_name() == "array"


def test_form_datatype():
    array = Array(shape=(12, 34))
    assert array.form_datatype() == "array<2>{real}"


def test_init():
    attrs = {"attr1": 1}
    array = Array(shape=(3,), dtype=np.float32, fill_val=42, attrs=attrs)
    assert (array.nda == np.full((3,), 42, np.float32)).all()
    assert array.attrs == attrs | {"datatype": "array<1>{real}"}


def test_ak_init():
    orig = ak.with_parameter([1, 2, 3, 4], "units", "mm")
    assert ak.parameters(orig)["units"] == "mm"

    array = Array(orig)
    assert isinstance(array.nda, np.ndarray)
    assert array.attrs["units"] == "mm"
    assert (array.nda == orig.to_numpy()).all()

    # check that this is a view
    array.nda[1] = -1
    assert orig[1] == -1


def test_resize_and_capacity():
    array = Array(nda=np.array([1, 2, 3, 4]))
    assert array.get_capacity() == 4

    array.resize(3)
    assert array.get_capacity() == 4
    assert (array.nda == np.array([1, 2, 3])).all()

    array.resize(5)
    assert array.get_capacity() >= 5

    array.clear(trim=True)
    assert array.get_capacity() == 0
    assert len(array) == 0


def test_insert():
    a = Array(np.array([1, 2, 3, 4]))
    a.insert(2, [-1, -1])
    assert a == Array([1, 2, -1, -1, 3, 4])

    with pytest.raises(IndexError):
        a.insert(10, 10)


def test_append():
    a = Array(np.array([1, 2, 3, 4]))
    a.append(-1)
    assert a == Array([1, 2, 3, 4, -1])


def test_replace():
    a = Array(np.array([1, 2, 3, 4]))
    a.replace(2, -1)
    assert a == Array([1, 2, -1, 4])

    with pytest.raises(IndexError):
        a.replace(10, 10)


def test_view():
    a = Array(np.array([1, 2, 3, 4]), attrs={"units": "m"})

    v = a.view_as("np", with_units=True)
    assert isinstance(v, pint.Quantity)
    assert v.u == "meter"
    assert np.array_equal(v.m, a.nda)

    v = a.view_as("np", with_units=False)
    assert isinstance(v, np.ndarray)

    v = a.view_as("pd", with_units=True)
    assert isinstance(v, pd.Series)
    assert v.dtype == "meter"

    v = a.view_as("pd", with_units=False)
    assert v.dtype == "int64"

    v = a.view_as("ak", with_units=False)
    assert isinstance(v, ak.Array)

    v = a.view_as("ak", with_units=True)
    assert isinstance(v, ak.Array)
    assert ak.parameters(v) == {"units": "m"}


def test_pickle():
    obj = Array(nda=np.array([1, 2, 3, 4]))
    obj.attrs["attr1"] = 1

    ex = pickle.loads(pickle.dumps(obj))
    assert isinstance(ex, Array)
    assert ex.attrs["attr1"] == 1
    assert ex.attrs["datatype"] == obj.attrs["datatype"]
    assert np.all(ex.nda == np.array([1, 2, 3, 4]))


def test_string_array():
    array = Array(ak.Array(["e", "sticazzi", "non", "ce", "li", "metti?"]))
    assert array.dtype == "<U16"



"""
Implements a LEGEND Data Object representing a special struct of arrays of
equal length and corresponding utilities.
"""

from __future__ import annotations

import logging
from collections.abc import Mapping
from types import ModuleType
from typing import Any
from warnings import warn

import awkward as ak
import numexpr as ne
import numpy as np
import pandas as pd
from pandas.io.formats import format as fmt

import lgdo

from .array import Array
from .arrayofequalsizedarrays import ArrayOfEqualSizedArrays
from .lgdo import LGDO, LGDOCollection
from .scalar import Scalar
from .struct import Struct
from .vectorofvectors import VectorOfVectors

log = logging.getLogger(__name__)


class Table(Struct, LGDOCollection):
    """A special struct of arrays or subtable columns of equal length.

    Note
    ----
    If you write to a table and don't fill it up to its total size, be sure to
    resize it before passing to data processing functions, as they will call
    :meth:`__len__` to access valid data, which returns the ``size`` attribute.
    """

    def __new__(cls, *args, **kwargs):
        # allow for (un-)pickling LGDO objects.
        obj = super().__new__(cls, *args, **kwargs)
        obj.size = None
        return obj

    def __init__(
        self,
        col_dict: Mapping[str, LGDOCollection] | pd.DataFrame | ak.Array | None = None,
        size: int | None = None,
        attrs: Mapping[str, Any] | None = None,
    ) -> None:
        r"""
        Parameters
        ----------
        size
            sets the number of rows in the table. :class:`.Array`\ s in
            `col_dict will be resized to match size if both are not ``None``.
            If `size` is left as ``None``, the number of table rows is
            determined from the length of the first array in `col_dict`. If
            neither is provided, a default length of 1024 is used.
        col_dict
            instantiate this table using the supplied mapping of column names
            and array-like objects. Supported input types are: mapping of
            strings to LGDOCollections, :class:`pd.DataFrame` and :class:`ak.Array`.
        attrs
            A set of user attributes to be carried along with this LGDO.

        Notes
        -----
        - The :attr:`loc` attribute is initialized to 0.
        - No copy is performed, the objects are used directly (unless
          :class:`ak.Array` is provided).
        - If `size` is not ``None``, all arrays will be resized to match it.
        - If the arrays have different lengths, all will be resized to match the
          length of the first array.

        Warning
        -------
        This constructor has partial units support. It supports fishing `units`
        from Awkward Array parameters but not (yet) from e.g. NumPy+Pint
        arrays. In any case, the user can always attach units later by
        modifying the dictionary held by :attr:`attrs`.
        """
        if isinstance(col_dict, pd.DataFrame):
            col_dict = {k: Array(v) for k, v in col_dict.items()}

        if isinstance(col_dict, ak.Array):
            col_dict = _ak_to_lgdo_or_col_dict(col_dict)

        # call Struct constructor
        Struct.__init__(self, obj_dict=col_dict, attrs=attrs)
        # no need to call the LGDOCollection constructor, as we are calling the
        # Struct constructor already

        # if col_dict is not empty, set size according to it
        # if size is also supplied, resize all fields to match it
        # otherwise, warn if the supplied fields have varying size
        if col_dict is not None and len(col_dict) > 0:
            self.resize(new_size=size, do_warn=(size is None))

        # if no col_dict, just set the size
        else:
            self.size = size if size is not None else None

    def datatype_name(self) -> str:
        return "table"

    def __len__(self) -> int:
        """Provides ``__len__`` for this array-like class."""
        return self.size

    def reserve_capacity(self, capacity: int | Mapping[str | int]) -> None:
        """Set size (number of rows) of internal memory buffer

        Parameters
        ----------
        capacity
            new capacities for fields in table. If `int`, set all capacities
            to value; if `Mapping`, set capacity field-by-field.
        """
        if isinstance(capacity, int):
            for obj in self.values():
                obj.reserve_capacity(capacity)
        else:
            for field, cap in capacity.items():
                self[field].reserve_capacity(cap)

    def get_capacity(self) -> dict[str, int]:
        "Return mapping from field name to capacity"
        return {k: v.get_capacity() for k, v in self.items()}

    def trim_capacity(self) -> int:
        "Set capacity for each column to be minimum needed to support size"
        for v in self.values():
            v.trim_capacity()

    def resize(
        self, new_size: int | None = None, do_warn: bool = False, trim: bool = False
    ) -> None:
        """Resize all columns of the table

        Parameters
        ----------
        new_size
            new size of table. If ``None`` use size of first field found
        do_warn
            emit a warning if contents for any field must be resized. This
            is intended for use with ``new_size = None``
        trim
            call :meth:`trim_capacity` after resizing to conserve memory
        """
        # if new_size = None, use the size from the first field
        for field, obj in self.items():
            if new_size is None:
                new_size = len(obj)
            elif len(obj) != new_size:
                if do_warn:
                    log.warning(
                        f"warning: resizing field {field}"
                        f"with size {len(obj)} != {new_size}"
                    )
                if isinstance(obj, Table):
                    obj.resize(new_size, trim)
                else:
                    obj.resize(new_size, trim)
        self.size = new_size

    def insert(self, i: int, vals: Table | Mapping[str, Any]) -> None:
        """Insert new row(s) into table

        Parameters
        ----------
        i
            row at which to insert values
        vals
            values to add. Require same keys as table
        """
        new_size = None
        for k, ar in self.items():
            ar.insert(i, vals[k])
            if new_size is None:
                new_size = len(ar)
            elif new_size != len(ar):
                msg = "inserted vals must all have same length"
                raise ValueError(msg)
        self.size = new_size

    def add_field(
        self,
        name: str,
        obj: LGDOCollection | Mapping[str, LGDOCollection],
        use_obj_size: bool = False,
    ) -> None:
        """Add a field (column) to the table.

        Use the name "field" here to match the terminology used in
        :class:`.Struct`.

        Parameters
        ----------
        name
            key to use for field. Key can be nested (e.g. ``name1.name2`` or
            ``name1/name2``); this will navigate through the tree, creating
            new fields as needed
        obj
            object to add. Can be any :class:`.LGDOCollection`, or a mapping from names
            to LGDOCollections that will be converted to an LGDO :class:`.Table`.
            Size of ``obj`` should match size of this Table
        use_obj_size
            if ``True``, resize the table to match the length of `obj`.
        """
        if not isinstance(obj, LGDO):
            if isinstance(obj, Mapping):
                obj = Table(obj, size=self.size)
            else:
                msg = "cannot add field of type"
                raise TypeError(msg, type(obj).__name__)

        super().add_field(name, obj)

        if self.size is None:
            self.size = len(obj)

        # check / update sizes
        if self.size != len(obj):
            warn(
                f"warning: you are trying to add {name} with length {len(obj)} to a table with size {self.size} and data might be lost. \n"
                f"With 'use_obj_size' set to:\n"
                f"  - True, the table will be resized to length {len(obj)} by padding/clipping its columns.\n"
                f"  - False (default), object {name} will be padded/clipped to length {self.size}.",
                UserWarning,
                stacklevel=2,
            )
            new_size = len(obj) if use_obj_size else self.size
            self.resize(new_size=new_size)

    def replace(self, i: int, vals: Mapping | Table) -> None:
        for k, ar in self.items():
            ar.replace(i, vals[k])

    def add_column(
        self, name: str, obj: LGDOCollection, use_obj_size: bool = False
    ) -> None:
        """Alias for :meth:`.add_field` using table terminology 'column'."""
        self.add_field(name, obj, use_obj_size=use_obj_size)

    def remove_column(self, name: str, delete: bool = False) -> None:
        """Alias for :meth:`Struct.remove_field` using table terminology 'column'."""
        super().remove_field(name, delete)

    def join(
        self,
        other_table: Table,
        cols: list[str] | None = None,
        keep_mine: bool = False,
        prefix: str = "",
        suffix: str = "",
        do_warn: bool = True,
    ) -> None:
        """Add the columns of another table to this table.

        Notes
        -----
        Following the join, both tables have access to `other_table`'s fields
        (but `other_table` doesn't have access to this table's fields). No
        memory is allocated in this process. `other_table` can go out of scope
        and this table will retain access to the joined data.

        Parameters
        ----------
        other_table
            the table whose columns are to be joined into this table.
        cols
            a list of names of columns from `other_table` to be joined into
            this table.
        keep_mine
            if there is a column name conflict, keep this Tables col if
            ``True``, or joined Table's column if ``False`` (default).
        prefix
            prepend to joined column names; can be used to avoid name conflicts.
        suffix
            append to joined column names; can be used to avoid name conflicts.
        do_warn
            set to ``False`` to turn off warnings associated with mismatched
            `loc` parameter or :meth:`add_column` warnings.
        """
        if len(other_table) != len(self) and do_warn:
            log.warning(
                f"len(other_table) ({len(other_table)}) != len(self) ({len(self)})"
            )
        if cols is None:
            cols = other_table.keys()
        for name in cols:
            joined_name = f"{prefix}{name}{suffix}"
            if joined_name in self and keep_mine:
                continue
            self.add_column(joined_name, other_table[name])

    def get_dataframe(
        self,
        cols: list[str] | None = None,
        copy: bool = False,  # noqa: ARG002
        prefix: str = "",
    ) -> pd.DataFrame:
        """Get a :class:`pandas.DataFrame` from the data in the table.

        Warning
        -------
        This method is deprecated. Use :meth:`.view_as` to view the table as a
        Pandas dataframe.

        Notes
        -----
        The requested data must be array-like, with the ``nda`` attribute.

        Parameters
        ----------
        cols
            a list of column names specifying the subset of the table's columns
            to be added to the dataframe.
        copy
            When ``True``, the dataframe allocates new memory and copies data
            into it. Otherwise, the raw ``nda``'s from the table are used directly.
        prefix
            The prefix to be added to the column names. Used when recursively getting the
            dataframe of a Table inside this Table
        """
        warn(
            "`get_dataframe` is deprecated and will be removed in a future release. "
            "Instead use `view_as` to get the Table data as a pandas dataframe "
            "or awkward Array. ",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.view_as(library="pd", cols=cols, prefix=prefix)

    def flatten(self, _prefix="") -> Table:
        """Flatten the table, if nested.

        Returns a new :class:`Table` (that references, not copies, the existing
        columns) with columns in nested tables being moved to the first level
        (and renamed appropriately).

        Examples
        --------
        >>> repr(tbl)
        "Table(dict={'a': Array([1 2 3], attrs={'datatype': 'array<1>{real}'}), 'tbl': Table(dict={'b': Array([4 5 6], attrs={'datatype': 'array<1>{real}'}), 'tbl1': Table(dict={'z': Array([9 9 9], attrs={'datatype': 'array<1>{real}'})}, attrs={'datatype': 'table{z}'})}, attrs={'datatype': 'table{b,tbl1}'})}, attrs={'datatype': 'table{a,tbl}'})"
        >>> tbl.flatten().keys()
        dict_keys(['a', 'tbl__b', 'tbl__tbl1__z'])
        """
        flat_table = Table(size=self.size)
        for key, obj in self.items():
            if isinstance(obj, Table):
                flat_table.join(obj.flatten(_prefix=f"{_prefix}{key}__"))
            else:
                flat_table.add_column(_prefix + key, obj)

        return flat_table

    def eval(
        self,
        expr: str,
        parameters: Mapping[str, str] | None = None,
        modules: Mapping[str, ModuleType] | None = None,
        with_units: bool = False,
        library: str | None = None,
    ) -> LGDO:
        """Apply column operations to the table and return a new LGDO.

        Internally uses :func:`numexpr.evaluate` if dealing with columns
        representable as NumPy arrays or :func:`eval` if
        :class:`.VectorOfVectors` are involved. In the latter case, the VoV
        columns are viewed as :class:`ak.Array` and the respective routines are
        therefore available.

        To columns nested in subtables can be accessed by scoping with two
        underscores (``__``). For example: ::

          tbl.eval("a + tbl2__b")

        computes the sum of column `a` and column `b` in the subtable `tbl2`.

        Parameters
        ----------
        expr
            if the expression only involves non-:class:`.VectorOfVectors`
            columns, the syntax is the one supported by
            :func:`numexpr.evaluate` (see `here
            <https://numexpr.readthedocs.io/projects/NumExpr3/en/latest/index.html>`_
            for documentation). Note: because of internal limitations,
            reduction operations must appear the last in the stack. If at least
            one considered column is a :class:`.VectorOfVectors`, plain
            :func:`eval` is used and :class:`ak.Array` transforms can be used
            through the ``ak.`` prefix. (NumPy functions are analogously
            accessible through ``np.``). See also examples below.
        parameters
            a dictionary of function parameters. Passed to
            :func:`numexpr.evaluate`` as `local_dict` argument or to
            :func:`eval` as `locals` argument.
        modules
            a dictionary of additional modules used by the expression. If this
            is not `None` then :func:`eval`is used and the expression can
            depend on any modules from this dictionary in addition to awkward
            and numpy. These are passed to :func:`eval` as `globals` argument.
        with_units
            attach units to the columns as in :meth:`LGDO.view_as`.
        library
            library to convert the columns to with :meth:`LGDO.view_as`,
            supported libraries are ``np``, ``ak`` or ``lgdo`` (pass in directly
            the unconverted LGDO objects).

        Examples
        --------
        >>> import lgdo
        >>> tbl = lgdo.Table(
        ...   col_dict={
        ...     "a": lgdo.Array([1, 2, 3]),
        ...     "b": lgdo.VectorOfVectors([[5], [6, 7], [8, 9, 0]]),
        ...   }
        ... )
        >>> print(tbl.eval("a + b"))
        [[6],
         [8 9],
         [11 12  3],
        ]
        >>> print(tbl.eval("np.sum(a) + ak.sum(b)"))
        41
        """
        if parameters is None:
            parameters = {}

        # get the valid python variable names in the expression
        c = compile(expr, "0vbb is real!", "eval")

        # make a dictionary of low-level objects (numpy or awkward)
        # for later computation
        flat_self = self.flatten()
        self_unwrap = {}
        has_only_np = False
        for obj in c.co_names:
            if obj in flat_self:
                # use the user-selected library or use the np/ak default depending on the type.
                col_library = library or "np"
                if library is None and isinstance(flat_self[obj], VectorOfVectors):
                    col_library = "ak"
                has_only_np = has_only_np or col_library == "np"

                if col_library == "lgdo":
                    self_unwrap[obj] = flat_self[obj]
                else:
                    self_unwrap[obj] = flat_self[obj].view_as(
                        col_library, with_units=with_units
                    )

        msg = f"evaluating {expr!r} with locals={(self_unwrap | parameters)} and {has_only_np=}"
        log.debug(msg)

        def _make_lgdo(data):
            if data.ndim == 0:
                return Scalar(data.item())
            if data.ndim == 1:
                return Array(data)
            if data.ndim == 2:
                return ArrayOfEqualSizedArrays(nda=data)

            msg = (
                f"evaluation resulted in {data.ndim}-dimensional data, "
                "I don't know which LGDO this corresponds to"
            )
            raise RuntimeError(msg)

        # use numexpr if we are only dealing with numpy data types (and no global dictionary)
        if has_only_np and modules is None:
            try:
                out_data = ne.evaluate(
                    expr,
                    local_dict=(self_unwrap | parameters),
                )

                msg = f"...the result is {out_data!r}"
                log.debug(msg)

                # need to convert back to LGDO
                # np.evaluate should always return a numpy thing?
                return _make_lgdo(out_data)

            except Exception:
                msg = (
                    f"Warning {expr} could not be evaluated with numexpr probably "
                    "due to some not allowed characters, trying with eval()."
                )
                log.debug(msg)

        # resort to good ol' eval()
        globs = {"ak": ak, "np": np, "lgdo": lgdo}
        if modules is not None:
            globs = globs | modules

        out_data = eval(expr, globs, (self_unwrap | parameters))

        msg = f"...the result is {out_data!r}"
        log.debug(msg)

        # need to convert back to LGDO
        if isinstance(out_data, ak.Array):
            if out_data.ndim == 1:
                return Array(out_data.to_numpy())
            return VectorOfVectors(out_data)

        # modules can still produce numpy array
        if isinstance(out_data, np.ndarray):
            return _make_lgdo(out_data)

        if np.isscalar(out_data):
            return Scalar(out_data)

        # if out_data is already an LGDO just return it
        if isinstance(out_data, LGDO):
            return out_data

        msg = (
            f"evaluation resulted in a {type(out_data)} object, "
            "I don't know which LGDO this corresponds to"
        )
        raise RuntimeError(msg)

    def __str__(self):
        opts = fmt.get_dataframe_repr_params()
        opts["show_dimensions"] = False
        opts["index"] = False

        try:
            string = self.view_as("pd").to_string(**opts)
        except ValueError:
            string = "Cannot print Table with VectorOfVectors yet!"

        string += "\n"
        for k, v in self.items():
            attrs = v.getattrs()
            if attrs:
                string += f"\nwith attrs['{k}']={attrs}"

        attrs = self.getattrs()
        if attrs:
            string += f"\nwith attrs={attrs}"

        return string

    def view_as(
        self,
        library: str,
        with_units: bool = False,
        cols: list[str] | None = None,
        prefix: str = "",
    ) -> pd.DataFrame | np.NDArray | ak.Array:
        r"""View the Table data as a third-party format data structure.

        This is typically a zero-copy or nearly zero-copy operation.

        Supported third-party formats are:

        - ``pd``: returns a :class:`pandas.DataFrame`
        - ``ak``: returns an :class:`ak.Array` (record type)

        Notes
        -----
        Conversion to Awkward array only works when the key is a string.

        Parameters
        ----------
        library
            format of the returned data view.
        with_units
            forward physical units to the output data.
        cols
            a list of column names specifying the subset of the table's columns
            to be added to the data view structure.
        prefix
            The prefix to be added to the column names. Used when recursively
            getting the dataframe of a :class:`Table` inside this
            :class:`Table`.

        See Also
        --------
        .LGDO.view_as
        """
        if cols is None:
            cols = self.keys()

        if library == "pd":
            df = {}

            for col in cols:
                data = self[col]

                if isinstance(data, Table):
                    log.debug(f"viewing Table {col=!r} recursively")

                    tmp_df = data.view_as(
                        "pd", with_units=with_units, prefix=f"{prefix}{col}_"
                    )
                    for k, v in tmp_df.items():
                        df[k] = v

                else:
                    log.debug(
                        f"viewing {type(data).__name__} column {col!r} as Pandas Series"
                    )
                    df[f"{prefix}{col}"] = data.view_as("pd", with_units=with_units)

            return pd.DataFrame(df, copy=False)

        if library == "np":
            msg = f"Format {library!r} is not supported for Tables."
            raise TypeError(msg)

        if library == "ak":
            # NOTE: passing the Table directly (which inherits from a dict) makes it
            # somehow really slow. ak.Array can unroll dicts just fine, but then uses
            # ak.from_iter to convert the arrays, instead of a zero-copy operation.
            return ak.Array(
                {col: self[col].view_as("ak", with_units=with_units) for col in cols}
            )

        msg = f"{library!r} is not a supported third-party format."
        raise TypeError(msg)


def _ak_to_lgdo_or_col_dict(array: ak.Array):
    if isinstance(array.type.content, ak.types.RecordType):
        return {field: _ak_to_lgdo_or_col_dict(array[field]) for field in array.fields}

    # be smart and just use Array when it makes sense
    if (
        isinstance(array.type.content, ak.types.NumpyType)
        or array.type.content.parameters.get("__array__") == "string"
    ):
        return Array(array)

    # otherwise fallback to VoV
    return VectorOfVectors(array)



"""
Implements a LEGEND Data Object representing an array of equal-sized arrays and
corresponding utilities.
"""

from __future__ import annotations

from collections.abc import Iterator
from typing import Any

import awkward as ak
import numpy as np
import pandas as pd

from .. import utils
from . import vectorofvectors as vov
from .array import Array


class ArrayOfEqualSizedArrays(Array):
    """An array of equal-sized arrays.

    Arrays of equal size within a file but could be different from application
    to application. Canonical example: array of same-length waveforms.
    """

    def __init__(
        self,
        dims: tuple[int, ...] | None = None,
        nda: np.ndarray = None,
        shape: tuple[int, ...] = (),
        dtype: np.dtype = None,
        fill_val: int | float | None = None,
        attrs: dict[str, Any] | None = None,
    ) -> None:
        """
        Parameters
        ----------
        dims
            specifies the dimensions required for building the
            :class:`ArrayOfEqualSizedArrays`' `datatype` attribute.
        nda
            An :class:`numpy.ndarray` to be used for this object's internal
            array. Note: the array is used directly, not copied. If not
            supplied, internal memory is newly allocated based on the `shape`
            and `dtype` arguments.
        shape
            A NumPy-format shape specification for shape of the internal
            array. Required if `nda` is ``None``, otherwise unused.
        dtype
            Specifies the type of the data in the array. Required if `nda` is
            ``None``, otherwise unused.
        fill_val
            If ``None``, memory is allocated without initialization. Otherwise,
            the array is allocated with all elements set to the corresponding
            fill value. If `nda` is not ``None``, this parameter is ignored.
        attrs
            A set of user attributes to be carried along with this LGDO.

        Notes
        -----
        If shape is not "1D array of arrays of shape given by axes 1-N" (of
        `nda`) then specify the dimensionality split in the constructor.

        See Also
        --------
        :class:`.Array`
        """
        if dims is None:
            # If no dims are provided, assume that it's a 1D Array of (N-1)-D Arrays
            if nda is None:
                s = shape
            else:
                if not isinstance(nda, np.ndarray):
                    nda = np.array(nda)
                s = nda.shape
            self.dims = (1, len(s) - 1)
        else:
            self.dims = dims
        super().__init__(
            nda=nda, shape=shape, dtype=dtype, fill_val=fill_val, attrs=attrs
        )

    def datatype_name(self) -> str:
        return "array_of_equalsized_arrays"

    def form_datatype(self) -> str:
        dt = self.datatype_name()
        nd = str(len(self.nda.shape))
        if self.dims is not None:
            nd = ",".join([str(i) for i in self.dims])
        et = utils.get_element_type(self)
        return dt + "<" + nd + ">{" + et + "}"

    def __len__(self) -> int:
        return len(self.nda)

    def __iter__(self) -> Iterator[np.array]:
        return self.nda.__iter__()

    def __next__(self) -> np.ndarray:
        return self.nda.__next__()

    def to_vov(self, cumulative_length: np.ndarray = None) -> vov.VectorOfVectors:
        """Convert (and eventually resize) to :class:`.vectorofvectors.VectorOfVectors`.

        Parameters
        ----------
        cumulative_length
            cumulative length array of the output vector of vectors. Each
            vector in the output is filled with values found in the
            :class:`ArrayOfEqualSizedArrays`, starting from the first index. if
            ``None``, use all of the original 2D array and make vectors of
            equal size.
        """
        attrs = self.getattrs()

        if cumulative_length is None:
            return vov.VectorOfVectors(
                flattened_data=self.nda.flatten(),
                cumulative_length=(np.arange(self.nda.shape[0], dtype="uint32") + 1)
                * self.nda.shape[1],
                attrs=attrs,
            )

        if not isinstance(cumulative_length, np.ndarray):
            cumulative_length = np.array(cumulative_length)

        flattened_data = self.nda[
            np.arange(self.nda.shape[1])
            < np.diff(cumulative_length, prepend=0)[:, None]
        ]

        return vov.VectorOfVectors(
            flattened_data=flattened_data,
            cumulative_length=cumulative_length,
            attrs=attrs,
        )

    def view_as(
        self, library: str, with_units: bool = False
    ) -> pd.DataFrame | np.NDArray | ak.Array:
        """View the array as a third-party format data structure.

        See Also
        --------
        .LGDO.view_as
        """
        return super().view_as(library, with_units=with_units)



from __future__ import annotations

import logging
import pickle
from pathlib import Path

import h5py
import numpy as np
import pytest

from lgdo import lh5, types
from lgdo.lh5.exceptions import LH5DecodeError, LH5EncodeError


def test_tools(tmptestdir):
    thefile = tmptestdir / "already-open-file-tools.lh5"
    lh5.write(types.Struct({}), "/empty", thefile)

    for mode in ("r", "r+", "w", "a"):
        f = h5py.File(thefile, mode)
        with pytest.raises(LH5DecodeError):
            lh5.ls(thefile)
        with pytest.raises(LH5DecodeError):
            lh5.show(thefile)
        f.close()


def test_interact_with_already_open_files(tmptestdir):
    thefile = tmptestdir / "already-open-file.lh5"
    lh5.write(types.Struct({}), "/empty", thefile)

    assert thefile.exists()

    f = h5py.File(thefile, "r")
    with pytest.raises(LH5EncodeError):
        lh5.write(types.Struct({}), "/empty2", thefile)
    f.close()

    for mode in ("r", "r+", "w", "a"):
        f = h5py.File(thefile, mode)
        with pytest.raises(LH5DecodeError):
            lh5.read("/empty", thefile)
        f.close()


def test_write_safe(tmptestdir):
    # write_safe should create new file
    struct = types.Struct()
    struct.add_field("scalar", types.Scalar(value=10, attrs={"sth": 1}))
    lh5.write(
        struct,
        "struct",
        f"{tmptestdir}/tmp-pygama-write_safe_exception.lh5",
        group="/data",
        start_row=1,
        n_rows=3,
        wo_mode="w",
    )

    # write_safe should not allow writing to existing dataset
    with pytest.raises(lh5.exceptions.LH5EncodeError):
        lh5.write(
            struct,
            "struct",
            f"{tmptestdir}/tmp-pygama-write_safe_exception.lh5",
            group="/data",
            start_row=1,
            n_rows=3,
            wo_mode="w",
        )


def test_open_non_existing_file(tmptestdir):
    with pytest.raises(LH5DecodeError):
        lh5.read("boh", tmptestdir / "beh.lh5")


def test_open_non_existing_dataset(tmptestdir):
    thefile = tmptestdir / "dummy-file.lh5"
    lh5.write(types.Struct({}), "/empty", thefile)

    with pytest.raises(LH5DecodeError):
        lh5.read("/boh", thefile)


def test_write_cleanup_on_error(tmptestdir):
    outfile = tmptestdir / "cleanup_error.lh5"
    h = types.Histogram(np.array([[1]]), (np.array([0, 1]), np.array([0, 1])))

    lh5.write(h, "hist", outfile, wo_mode="overwrite_file")
    with pytest.raises(lh5.exceptions.LH5EncodeError):
        lh5.write(h, "hist", outfile, wo_mode="append")

    # file should be closed so we can open it again
    assert lh5.read("hist", outfile) is not None


def test_write_append_struct(tmptestdir):
    outfile = tmptestdir / "test-write-append-struct.lh5"
    st = types.Struct({"arr1": types.Table({"a": types.Array([1, 2, 3])})})
    lh5.write(st, "struct", outfile, wo_mode="of")
    st2 = types.Struct({"arr2": types.Table({"a": types.Array([1, 2, 3])})})
    lh5.write(st2, "struct", outfile, wo_mode="ac")

    # test error when appending existing field
    with pytest.raises(lh5.exceptions.LH5EncodeError):
        lh5.write(
            types.Struct({"arr2": types.Array([4, 5, 6])}),
            "struct",
            outfile,
            wo_mode="ac",
        )

    # error if appending to object of different type
    with pytest.raises(lh5.exceptions.LH5EncodeError):
        lh5.write(
            types.Struct({"arr2": types.Array([4, 5, 6])}),
            "struct",
            outfile,
            wo_mode="ac",
        )

    lh5.write(
        types.Table({"arr1": types.Array([1, 2, 3])}), "struct", outfile, wo_mode="of"
    )

    # error if appending to object of different type
    with pytest.raises(lh5.exceptions.LH5EncodeError):
        lh5.write(
            types.Table({"arr2": types.Array([4, 5, 6, 7])}),
            "struct",
            outfile,
            wo_mode="ac",
        )


# Test that when we try to overwrite an existing column in a table we fail
def test_write_object_append_column(tmptestdir):
    # Try to append an array to a table
    if Path(f"{tmptestdir}/write_object_append_column_test.lh5").exists():
        Path(f"{tmptestdir}/write_object_append_column_test.lh5").unlink()

    array1 = types.Array(np.zeros(10))
    tb1 = types.Table(col_dict={"dset1`": types.Array(np.ones(10))})
    store = lh5.LH5Store()
    store.write(array1, "my_table", f"{tmptestdir}/write_object_append_column_test.lh5")
    with pytest.raises(lh5.exceptions.LH5EncodeError):
        store.write(
            tb1,
            "my_table",
            f"{tmptestdir}/write_object_append_column_test.lh5",
            wo_mode="append_column",
        )  # Now, try to append a column to an array

    # Try to append a table that has a same key as the old table
    if Path(f"{tmptestdir}/write_object_append_column_test.lh5").exists():
        Path(f"{tmptestdir}/write_object_append_column_test.lh5").unlink()

    tb1 = types.Table(
        col_dict={
            "dset1": types.Array(np.zeros(10)),
            "dset2": types.Array(np.zeros(10)),
        }
    )
    tb2 = types.Table(
        col_dict={"dset2": types.Array(np.ones(10))}
    )  # Same field name, different values
    store = lh5.LH5Store()
    store.write(tb1, "my_table", f"{tmptestdir}/write_object_append_column_test.lh5")
    with pytest.raises(lh5.exceptions.LH5EncodeError):
        store.write(
            tb2,
            "my_table",
            f"{tmptestdir}/write_object_append_column_test.lh5",
            wo_mode="append_column",
        )  # Now, try to append a column with a same field

    # try appending a column that is larger than one that exists
    if Path(f"{tmptestdir}/write_object_append_column_test.lh5").exists():
        Path(f"{tmptestdir}/write_object_append_column_test.lh5").unlink()

    tb1 = types.Table(col_dict={"dset1": types.Array(np.zeros(10))})
    tb2 = types.Table(
        col_dict={"dset2": types.Array(np.ones(20))}
    )  # different field name, different size
    store = lh5.LH5Store()
    store.write(tb1, "my_table", f"{tmptestdir}/write_object_append_column_test.lh5")
    with pytest.raises(lh5.exceptions.LH5EncodeError):
        store.write(
            tb2,
            "my_table",
            f"{tmptestdir}/write_object_append_column_test.lh5",
            wo_mode="append_column",
        )  # Now, try to append a column with a different field size


# Test writing and reading histograms.
def test_write_histogram(caplog, tmptestdir):
    caplog.set_level(logging.DEBUG)
    caplog.clear()

    # Start with an types.Histogram
    if Path(f"{tmptestdir}/write_histogram_test.lh5").exists():
        Path(f"{tmptestdir}/write_histogram_test.lh5").unlink()

    h1 = types.Histogram(
        np.array([[1, 1], [1, 1]]), (np.array([0, 1, 2]), np.array([2.1, 2.2, 2.3]))
    )
    h2 = types.Histogram(
        np.array([[10, 10], [10, 10]]),
        (np.array([2, 3, 4]), np.array([5, 6, 7])),
        isdensity=True,
    )
    h2.binning[0]["binedges"].attrs["units"] = "ns"

    # Same field name, different values
    store = lh5.LH5Store()
    # "appending" to a non-existing histogram should work.
    store.write(
        h1,
        "my_histogram",
        f"{tmptestdir}/write_histogram_test.lh5",
        group="my_group",
        wo_mode="append",
    )
    with pytest.raises(lh5.exceptions.LH5EncodeError):
        # appending to an existing histogram should not work.
        store.write(
            h1,
            "my_histogram",
            f"{tmptestdir}/write_histogram_test.lh5",
            group="my_group",
            wo_mode="append",
        )
    store.write(
        h2,
        "my_histogram",
        f"{tmptestdir}/write_histogram_test.lh5",
        wo_mode="overwrite",
        group="my_group",
    )

    # Now, check that writing with other modes throws.
    for disallowed_wo_mode in ["append", "append_column"]:
        with pytest.raises(lh5.exceptions.LH5EncodeError):
            store.write(
                h2,
                "my_histogram",
                f"{tmptestdir}/write_histogram_test.lh5",
                wo_mode=disallowed_wo_mode,
                group="my_group",
            )
    with pytest.raises(lh5.exceptions.LH5EncodeError):
        store.write(
            h2,
            "my_histogram",
            f"{tmptestdir}/write_histogram_test.lh5",
            wo_mode="overwrite",
            write_start=1,
            group="my_group",
        )


def test_pickle():
    # test (un-)pickling of LH5 exceptions; e.g. for multiprocessing use.

    ex = LH5EncodeError("message", "file", "group", "name")
    ex = pickle.loads(pickle.dumps(ex))
    assert isinstance(ex, LH5EncodeError)

    ex = LH5DecodeError("message", "file", "obj")
    ex = pickle.loads(pickle.dumps(ex))
    assert isinstance(ex, LH5DecodeError)
